/*****
 * xscreenshot.c
 * Take screenshot, with selectable area.
 * Copyright (C) James Budiono, 2014, 2023
 * License: GNU GPL Version 3 or later.
 *****/

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Xregion.h>
#include <X11/extensions/Xfixes.h>
#include <X11/extensions/shape.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include "pnglite.h"

#define DEFAULT_SCREENSHOT_PREFIX "xscreenshot"
#define DEFAULT_SCREENSHOT_TAKEN_MSG "Screenshot taken!"
#define MSG_FONTNAME "dejavu sans"
#define MSG_FONTSIZE "30"
#define MSG_DELAY 3
#define DEFAULT_PEN_WIDTH 1
#define DEFAULT_PEN_COLOUR "White"

// so we don't pass thousands of parameters each time
typedef struct
{
	Display *display;
	Window	root_window;
	int     screen;
	int     width, height;
	unsigned int hotkey_keycode, hotkey_keycode2;
	
	GC     pen;	
	Cursor grab_cursor;
	BOX    anchor; //x1,y1=anchor, x2,y2=current pointer
	int    select_in_progress;
	int    mouse_grabbed;

	int  save_to_file, save_to_clipboard;
	char *screenshot_path;
	char *screenshot_buf;
	size_t screenshot_bufsize;
	Atom XA_CLIPBOARD, XA_TARGETS, XA_MULTIPLE, XA_image_png;

	int screenshot_delay;

#ifdef USE_OVERLAY
	// overlay window
	Window overlay;	
	Pixmap overlay_content, overlay_shapemask;
	XVisualInfo overlay_visinfo;
	Colormap cmap;
	Atom wm_delete_window;
	Atom wm_protocols;	
	GC content_blank;
	GC bitmap_white, bitmap_black;
	int visible;
#endif

} ProgState;



////////////// SCREEN CAPTURE ROUTINE //////////////

// get shift count based on mask, assumes LSB
int get_shift (int mask)
{
	int n=0, test=1;
	while (n < sizeof(mask)*8) {
		if (mask & test) return n;
		n++; test<<=1;
	}
	return 0;
}

// get 24-bit rgb data from XImage, ZPixmap format
// only supports direct/truecolor 15/16/24 bit, assumes LSB
unsigned char *extract_rgb24_from_ximage (XImage *image)
{
	unsigned char *out=NULL;
	int x,y,index;
	int width, height;
	unsigned long pix;
	
	int red_shift, green_shift, blue_shift;
	int red_mask, green_mask, blue_mask;
	double red_factor=1.0, green_factor=1.0, blue_factor=1.0;
	
	width       = image->width;
	height      = image->height;
	red_mask    = image->red_mask;
	green_mask  = image->green_mask;
	blue_mask   = image->blue_mask;
	red_shift   = get_shift (red_mask);
	green_shift = get_shift (green_mask);
	blue_shift  = get_shift (blue_mask);
	
	switch (image->depth) {
		case 15:
		case 16:
			red_factor   = 255.0/(red_mask   >> red_shift);
			green_factor = 255.0/(green_mask >> green_shift);
			blue_factor  = 255.0/(blue_mask  >> blue_shift);		
		case 24:
		case 32:
			break;
		default:
			fprintf (stderr, "Unsupported bit depth: %d.\n", image->depth);
			return NULL;
	}
	
	out = malloc (width * height * 3); // 24-bit RGB (3 bytes per pixel)	
	for (index=0, y=0; y < height; y++)
	{
		for (x=0; x < width; x++)
		{
			pix = XGetPixel (image, x,y);
			out[index++] = ((pix & red_mask)   >> red_shift) * red_factor;
			out[index++] = ((pix & green_mask) >> green_shift) * green_factor;
			out[index++] = ((pix & blue_mask)  >> blue_shift) * blue_factor;
		}
	}
	return out;
}

// overlay the mouse cursor on top of given image.
void paint_mouse_cursor(ProgState *st, BOX rect, unsigned char *image) {
	/* Check if the XFIXES extension is present */
	int event, error;
	if (XFixesQueryExtension(st->display, &event, &error))
	{
		XFixesCursorImage *cur = XFixesGetCursorImage(st->display);
		
		int sx = (cur->x - cur->xhot), sy = (cur->y - cur->yhot);
		int x,y,dx,dy, cx,cy;
		for (y = sy, dy=cur->height, cy=0;  dy>0; dy--, y++, cy++) {
			for (x = sx, dx=cur->width, cx=0; dx> 0; dx--, x++, cx++) {
				
				if (x>=rect.x1 && y>=rect.y1 && x<rect.x2 && y<rect.y2) {
					unsigned long cur_pixel = cur->pixels [ cy*cur->width + cx ];
					int r = (cur_pixel >> 0) & 0xff;
					int g = (cur_pixel >> 8) & 0xff;
					int b = (cur_pixel >> 16) & 0xff;
					int a = (cur_pixel >> 24) & 0xff;
					
					int img_ofs = ((y-rect.y1)*rect.x2 + (x-rect.x1)) * 3; // 24-bit RGB - 3 bytes per pixel
					if (a == 255) {
						image [ img_ofs +0 ] = r;
						image [ img_ofs +1 ] = g;
						image [ img_ofs +2 ] = b;
					} else {
						image [ img_ofs +0 ] = r + (image [ img_ofs +0 ] * (255 - a) + 255 / 2) / 255;
						image [ img_ofs +1 ] = g + (image [ img_ofs +1 ] * (255 - a) + 255 / 2) / 255;
						image [ img_ofs +2 ] = b + (image [ img_ofs +2 ] * (255 - a) + 255 / 2) / 255;
					}
				}
				
			}
		}
	}
}


// This must be freed when done
char *get_save_fname(ProgState *st)
{
	char *fname;
	char ts[64];
	
	time_t now = time(NULL);
	struct tm *now2 = localtime(&now);
	strftime(ts, sizeof(ts), "%Y%m%dT%H%M%S", now2);
	asprintf (&fname, "%s-%s.png", st->screenshot_path, ts);
	return fname;
}

// helper routine for png_open_write
unsigned fmemwrite(void* input, size_t size, size_t numel, void* user_pointer) {
	FILE *fmem = (FILE *) user_pointer;
	return fwrite(input, size, numel, fmem);
}
void clear_clipboard_data(ProgState *st) {				
	if (st->screenshot_buf) {
		free(st->screenshot_buf);
		st->screenshot_buf = 0;
		st->screenshot_bufsize = 0;
	}
}

// take screenshot using pnglite
void take_screenshot (ProgState *st, BOX rect, int with_mouse) // x2,y2 is width, height
{
	XImage *image;
	png_t png;
	unsigned char *buf;
	char *fname=NULL;
	FILE *fmem;
	
	// sanity check - if either height or width is zero, abort
	if (!rect.x2 || !rect.y2) return;

	// make sure all previous operations are completed
	XSync(st->display, False);

	// do actual screenshot
	image = XGetImage (
		st->display, st->root_window, rect.x1, rect.y1, rect.x2, rect.y2, 
		AllPlanes, ZPixmap
	);
	buf = extract_rgb24_from_ximage (image);
	if (with_mouse) 
		paint_mouse_cursor (st, rect, buf);
	
	// save it
	if (buf) {
		if (st->save_to_file) {
			fname = get_save_fname(st);
			png_open_file_write (&png, fname);
			png_set_data (&png, image->width, image->height, 8, PNG_TRUECOLOR, buf); //8 bits per channel.
			png_close_file (&png);
			fprintf(stderr, "saved screenshot to %s\n", fname);
			free (fname);
		}
		
		if (st->save_to_clipboard) {
			clear_clipboard_data(st);
			fmem = open_memstream(&st->screenshot_buf, &st->screenshot_bufsize);
			png_open_write(&png, &fmemwrite, fmem);
			png_set_data (&png, image->width, image->height, 8, PNG_TRUECOLOR, buf); //8 bits per channel.
			fclose(fmem);
			
			if (st->screenshot_buf && st->screenshot_bufsize) {
				// indicate we want to put data in clipboard
				XSetSelectionOwner(st->display, st->XA_CLIPBOARD, st->root_window, CurrentTime);
				fprintf(stderr, "saved screenshot to clipboard\n");
			}
		}
		
		free (buf);	
	}
	
	XDestroyImage (image);
	fprintf (stderr, DEFAULT_SCREENSHOT_TAKEN_MSG "\n");
}

///////////////// END OF SCREEN CAPTURE ROUTINE ///////////////





///////////////// OVERLAY WINDOW MANAGEMENT ///////////////////

#ifdef USE_OVERLAY
// create the overlay window we will use as canvas
void create_overlay_window (ProgState *st, Window *parent) 
{
	XSetWindowAttributes xswa;
	unsigned long attrmask=0;

	st->overlay_visinfo.visual = DefaultVisual (st->display, st->screen);
	st->overlay_visinfo.depth  = DefaultDepth (st->display, st->screen);
	st->cmap = DefaultColormap(st->display, st->screen);
	xswa.background_pixel = WhitePixel(st->display, st->screen);
	attrmask |= CWBackPixel;

	xswa.save_under = True;
	xswa.backing_store = NotUseful;
	xswa.override_redirect = True;
	xswa.cursor = None;
	attrmask |= CWBackingStore|CWSaveUnder|CWOverrideRedirect|CWCursor;
	
	st->overlay = XCreateWindow (st->display, *parent, 0,0, st->width,st->height, 0, 
	                     st->overlay_visinfo.depth, InputOutput, st->overlay_visinfo.visual, 
	                     attrmask, &xswa);
	                     
    XShapeCombineRectangles (st->display, st->overlay, ShapeInput, 0, 0, NULL, 0, ShapeSet, False);
	XSelectInput(st->display, st->overlay, StructureNotifyMask|ExposureMask);
    XSetWMProtocols(st->display, st->overlay, &st->wm_delete_window, 1);
    
	XMapWindow(st->display, st->overlay);
	XStoreName(st->display, st->overlay, "xscreenshot");
}

// clear canvas
void clear_drawing (ProgState *st)
{
	//fprintf (stderr, "clear drawing\n");
	XFillRectangle (st->display, st->overlay_content, st->content_blank, 0, 0, st->width, st->height);
	XFillRectangle (st->display, st->overlay_shapemask, st->bitmap_black, 0, 0, st->width, st->height);	
}

// update window content with latest drawing
void repaint (ProgState *st)
{
	// shape-combine has to go first
	XShapeCombineMask(st->display, st->overlay, ShapeBounding, 0, 0, st->overlay_shapemask, ShapeSet);	
	XCopyArea (st->display, st->overlay_content, st->overlay, st->content_blank, 0, 0, st->width, st->height, 0, 0);
}

// show or hide window
void show_window (ProgState *st, int visible)
{
	if (visible) {
		if (!st->visible) {
			//fprintf (stderr, "show window\n");
			XMapWindow(st->display, st->overlay);
			st->visible = visible;
		}
	} else {
		if (st->visible) {
			//fprintf (stderr, "hide window\n");
			XUnmapWindow(st->display, st->overlay);
			st->visible = visible;
		}		
	}
}

// create GC (pen) for selection box
GC create_pen (ProgState *st, Drawable d, char *colour, int width, XFontStruct *font)
{
	GC pen;
	XColor xc, xc2;
	XGCValues values;
	unsigned long gcmask;
	
	xc.pixel = 0; // transparent
	if (colour)
		XAllocNamedColor(st->display, st->cmap, colour, &xc, &xc2);
	
	/* create the pen to draw lines with */
	values.line_width = width;
	values.line_style = LineSolid;
	values.foreground = xc.pixel;
	values.cap_style = CapRound;
	values.join_style = JoinMiter;
	values.subwindow_mode = IncludeInferiors;
	gcmask = GCForeground | GCLineWidth | GCLineStyle | GCCapStyle | GCJoinStyle | GCSubwindowMode;
	if (font)
	{
		values.font = font->fid;
		gcmask |= GCFont;
	}
	return XCreateGC(st->display, d, gcmask, &values);
}

///////////////// END OF OVERLAY WINDOW MANAGEMENT ///////////////

#else
// create GC (pen) for selection box
GC create_pen (ProgState *st, Drawable d, int width, XFontStruct *font)
{
	XGCValues values;
	unsigned long gcmask;
	
	values.function = GXinvert;
	values.subwindow_mode = IncludeInferiors;
	values.line_width = width;
	gcmask = GCFunction | GCSubwindowMode | GCLineWidth;
	if (font)
	{
		values.font = font->fid;
		gcmask |= GCFont;
	}
    return XCreateGC (st->display, d, gcmask, &values);
}
#endif




/////////////////// HOTKEY SELECTION ETC /////////////////////


// enable notification when our hotkey is pressed
char *grab_hotkey (ProgState *st, char *hotkey)
{
	char *hotkey_name = hotkey;
	unsigned int keycode = 0, keycode2 = 0;
	KeySym keysym, keysym2;
	
	keysym = XStringToKeysym (hotkey);
	if (keysym != NoSymbol)
		keycode = XKeysymToKeycode (st->display, keysym);
	else {
		keycode = atoi(hotkey);
		if (keycode) {
			keysym2 = XKeycodeToKeysym(st->display, keycode, 0);
			keycode2 = XKeysymToKeycode (st->display, keysym2);
			hotkey_name = XKeysymToString (keysym2);
		}
	}
		
	if (keycode)
		XGrabKey (st->display, keycode, AnyModifier, st->root_window, 
		          False, GrabModeAsync, GrabModeAsync);
	if (keycode2 && keycode2 != keycode)
		XGrabKey (st->display, keycode2, AnyModifier, st->root_window, 
		          False, GrabModeAsync, GrabModeAsync);	
	if (!keycode && !keycode2) 
	{
		fprintf (stderr, "I don't understand %s, bailing out.\n", hotkey);
		exit(1);
	}
	st->hotkey_keycode = keycode;
	st->hotkey_keycode2 = keycode2;
	return hotkey_name;
}


// grap or ungrab the mouse pointer
int grab_ungrab_mouse (ProgState *st, int grab)
{
	if (grab)
	{
		//fprintf (stderr, "mouse grab\n");
		XGrabPointer (st->display, st->root_window, False, 
					  ButtonPressMask|ButtonReleaseMask|PointerMotionMask,
					  GrabModeAsync, GrabModeAsync, None, st->grab_cursor, CurrentTime);
		
	} else {
		//fprintf (stderr, "mouse release\n");
		XUngrabPointer (st->display, CurrentTime);
	}
	st->mouse_grabbed = grab;	
}


// draw a rectangle, x2,y2 is width, height
void draw_rect (ProgState *st, BOX rect)
{
	//fprintf (stderr, "%d, %d, %d, %d\n", rect.x1, rect.y1, rect.x2, rect.y2);
#ifdef USE_OVERLAY
	XDrawRectangle (
		st->display, st->overlay_content, st->pen, rect.x1, rect.y1, rect.x2, rect.y2
	);
	XDrawRectangle (
		st->display, st->overlay_shapemask, st->bitmap_white, rect.x1, rect.y1, rect.x2, rect.y2
	);
	repaint(st);
#else
	XDrawRectangle (
		st->display, st->root_window, st->pen, rect.x1, rect.y1, rect.x2, rect.y2
	);
#endif
}

// convert anchor points to a rect (x,y,width,height)
BOX convert_anchor_to_rect (BOX *anchor)
{
	BOX out;
	out.x1 = anchor->x1;
	if (anchor->x1 > anchor->x2) out.x1 = anchor->x2;
	out.y1 = anchor->y1;
	if (anchor->y1 > anchor->y2) out.y1 = anchor->y2;
	out.x2 = abs (anchor->x1 - anchor->x2);
	out.y2 = abs (anchor->y1 - anchor->y2);
	return out;
}

// convert anchor points to an area (x,y,width,height) 
// inclusive of the anchors themselves
BOX convert_anchor_to_area (BOX *anchor) {
	BOX out = convert_anchor_to_rect (anchor);
	out.x2++;
	out.y2++;
	return out;
}

// start selection
void start_selection (ProgState *st, int x, int y)
{
#ifdef USE_OVERLAY
	show_window(st, 1);
#endif
	st->anchor.x1 = st->anchor.x2 = x;
	st->anchor.y1 = st->anchor.y2 = y;
	st->select_in_progress = 1;
	draw_rect (st, convert_anchor_to_rect (&st->anchor));
}
// end active selection if currently in progress
void end_selection (ProgState *st)
{
	if (st->select_in_progress)
	{
		st->select_in_progress = 0;
#ifdef USE_OVERLAY
		clear_drawing(st);
		repaint(st);
		show_window(st, 0);
#else
		draw_rect (st, convert_anchor_to_rect (&st->anchor));
#endif		
	}
}
// draw selection box if selection is in progres
void draw_selection (ProgState *st, int x, int y)
{
	if (st->select_in_progress)
	{
#ifdef USE_OVERLAY
		clear_drawing(st);
#else
		draw_rect (st, convert_anchor_to_rect (&st->anchor));
#endif
		st->anchor.x2 = x;
		st->anchor.y2 = y;
		draw_rect (st, convert_anchor_to_rect (&st->anchor));
	}
}

// show notification on screen that capture has been completed
void notify(ProgState *st, char *msg, int x, int y, int delay) {
	int msglen;
	
	if (!msg) return; 
	msglen = strlen(msg);

#ifdef USE_OVERLAY
	show_window(st, 1);
	XDrawString(st->display, st->overlay_content, st->pen, x, y, msg, msglen);
	XDrawString(st->display, st->overlay_shapemask, st->bitmap_white, x, y, msg, msglen);
	XSync(st->display, False);
	repaint(st);
	XSync(st->display, False);
	sleep(delay);

	clear_drawing(st);
	repaint(st);	
	show_window(st, 0);
#else	
	XDrawString(st->display, st->root_window, st->pen, x, y, msg, msglen);
	XSync(st->display, False);
	sleep (delay);
	XDrawString(st->display, st->root_window, st->pen, x, y, msg, msglen);
	XSync(st->display, False);
#endif
}

// perform delayed screen capture
void delayed_capture(ProgState *st, int with_mouse)
{
	BOX rect = { .x1=0, .y1=0, .x2=st->width, .y2=st->height };

	fprintf (stderr, "screenshot in %d seconds\n", st->screenshot_delay);
	sleep (st->screenshot_delay);
	st->save_to_file = st->save_to_clipboard = 1; // save to both
	take_screenshot (st, rect, with_mouse);
	notify (st, DEFAULT_SCREENSHOT_TAKEN_MSG "\n", 0, atoi(MSG_FONTSIZE), MSG_DELAY);
}

// get the area of window under cursor
BOX get_window_area_under_cursor(ProgState *st) 
{
	BOX area;
	area.x1 = area.x2 = area.y1 = area.y2 = 0;
	
	Window root_return, child_return;
	int root_x_return, root_y_return;
	int win_x_return, win_y_return;
	unsigned int mask_return;	

	if (XQueryPointer(st->display, st->root_window, 
			&root_return, &child_return, // we only want this
			&root_x_return, &root_y_return, &win_x_return, &win_y_return, &mask_return) == True)
	{
		Window win = child_return;
		if (win == None) {
			win = root_return;
		} else {
			// raise it to make sure it is not obscured
			XRaiseWindow(st->display,win);
			XFlush(st->display);
			sleep(1); // sleep 1 to ensure display is updated
		}

        int x_return, y_return;
        unsigned int width_return, height_return;
        unsigned int border_width_return;
        unsigned int depth_return;
        
		if (XGetGeometry(st->display, win, &root_return, 
				&x_return, &y_return, &width_return, &height_return, // we only want this
				&border_width_return, &depth_return) == True)
		{
			// handle partial off-screen window - Jake
			if (x_return < 0) { width_return += x_return; x_return = 0; }
			if (y_return < 0) { height_return += y_return; y_return = 0; }
			if ( (x_return + width_return) >= st->width ) { width_return = st->width - x_return; }
			if ( (y_return + height_return) >= st->height ) { height_return = st->height - y_return; }			
			
			area.x1 = x_return; area.y1 = y_return;
			area.x2 = width_return; area.y2 = height_return;
		}
	}
	return area;
}

void toggle_selection_capture(ProgState *st) {
	if (st->mouse_grabbed)
	{
		if (st->select_in_progress) 
			end_selection (st);
		grab_ungrab_mouse (st, 0);
	}
	else
	{
		grab_ungrab_mouse (st, 1);
	}
}


const char *GetAtomName(Display* disp, Atom a)
{
	if(a == None)
		return "None";
	else
		return XGetAtomName(disp, a);
}
void handle_clipboard_request(ProgState *st, XEvent ev) {
	Atom targets[] = { st->XA_TARGETS, st->XA_MULTIPLE, st->XA_image_png };
	XEvent reply;
	
	// Extract data
	Window owner     = ev.xselectionrequest.owner;
	Atom selection   = ev.xselectionrequest.selection;
	Atom target      = ev.xselectionrequest.target;
	Atom property    = ev.xselectionrequest.property;
	Window requestor = ev.xselectionrequest.requestor;
	Time timestamp   = ev.xselectionrequest.time;
	Display* disp    = ev.xselection.display;

	// X should only send requests for the selections we own.
	// Since we own exaclty one (CLIPBOARD), we don't need to determine 
	// which selection that we need to handle

	// Construct reply event. We 'refuse' by default.
	reply.xselection.type      = SelectionNotify;
	//reply.xselection.serial     - filled in by server
	//reply.xselection.send_event - filled in by server
	//reply.xselection.display    - filled in by server
	reply.xselection.requestor = requestor;
	reply.xselection.selection = selection;
	reply.xselection.target    = target;
	reply.xselection.property  = None;   //This means refusal
	reply.xselection.time      = timestamp;
	
	if (target == st->XA_TARGETS)
	{
		XChangeProperty(st->display, requestor, property, XA_ATOM, 32,
						PropModeReplace, (const unsigned char *)targets, sizeof(targets)/sizeof(Atom));
		reply.xselection.property = property;
	}
	else if (target == st->XA_image_png && st->screenshot_buf && st->screenshot_bufsize)
	{
		fprintf(stderr,"sending image\n");
		reply.xselection.property = property;
		XChangeProperty(disp, requestor, property, target, 8, PropModeReplace,
						st->screenshot_buf, st->screenshot_bufsize);
						
	}
	else if (target == st->XA_MULTIPLE)
	{
		// The example I have does not support this request
		//fprintf(stderr, "MULTIPLE is not implemented yet.\n");
	}
	else
	{
		// Everything else is not supported
		//fprintf(stderr, "%s is not supported\n", GetAtomName(st->display, target));
	}

	// Reply
	XSendEvent(disp, ev.xselectionrequest.requestor, True, 0, &reply);
}





////////////////////// MAIN EVENT LOOP /////////////////////

int event_loop (ProgState *st)
{
	XEvent ev;
	int running=1;
	int state;
	BOX capture_area;
	
	while (running)
	{
next_event:
		XNextEvent(st->display, &ev);
		switch(ev.type)
		{
			case ButtonPress:
				if (ev.xbutton.button == Button1)
					start_selection (st, ev.xbutton.x, ev.xbutton.y);
				break;
				
			case ButtonRelease:
				switch (ev.xbutton.button)
				{
					case Button1:
						if (st->select_in_progress)
						{	// selection done
							end_selection (st);

							capture_area = convert_anchor_to_area (&st->anchor);
							if (capture_area.x2 <= 2 && capture_area.y2 <= 2)
								capture_area = get_window_area_under_cursor (st);

							take_screenshot (st, capture_area, 0);
							if (!st->hotkey_keycode) {
								grab_ungrab_mouse (st, 0);
								running=0;
							}
						}
						break;
						
					case Button3:
						if (st->select_in_progress)
						{
							end_selection (st);
						}
						else
						{	// leave if not running in multi shots mode
							if (!st->hotkey_keycode)
							{
								grab_ungrab_mouse (st, 0);
								running=0;
							}
						}
						
						break;
				}
				break;
				
			case MotionNotify:
				draw_selection (st, ev.xmotion.x, ev.xmotion.y);
				break;
			
			case KeyPress:
				if (ev.xkey.keycode == st->hotkey_keycode ||
				    ev.xkey.keycode == st->hotkey_keycode2 )
				{
					state=0;
					if (ev.xkey.state & ShiftMask) state ^= 1;
					if (ev.xkey.state & ControlMask) state ^= 2;
					if (ev.xkey.state & Mod1Mask) state ^= 4; // Alt
					switch (state)
					{
						case 0: // bare - region capture to file
							st->save_to_file = 1; st->save_to_clipboard = 0;
							toggle_selection_capture(st);
							break;

						case 2: // ctrl - region capture to clipboard
							st->save_to_file = 0; st->save_to_clipboard = 1;
							toggle_selection_capture(st);
							break;
														
						case 1: // shift - delayed fullscreen shot with mouse
							delayed_capture (st, 1);
							break;

						case 4: // alt - delayed fullscreen shot without mouse
							delayed_capture (st, 0);
							break;													

						case 3: // ctrl-shift - drop clipboard
							if (st->screenshot_buf) {
								fprintf(stderr, "dropped clipboard\n");
								clear_clipboard_data(st);
								XSetSelectionOwner(st->display, st->XA_CLIPBOARD, None, CurrentTime);
							}
							break;							
/*
						case 7: // alt-ctrl-shift - exit
							if (st->select_in_progress) 
								end_selection (st);
							if (st->mouse_grabbed)
								grab_ungrab_mouse (st, 0);
							running=0;
							break;
*/
					}
				}

#ifdef USE_OVERLAY
			// quit if our overlay window is deleted
			case ClientMessage:
				if (ev.xclient.message_type == st->wm_protocols && 
				   (Atom)ev.xclient.data.l[0] == st->wm_delete_window)
				   running=0;
				break;
#endif

			case SelectionClear:
				clear_clipboard_data(st);
				fprintf(stderr,"clipboard cleared\n");
				break;

			case SelectionRequest:
				handle_clipboard_request(st, ev);
				break;
		}
	}
	return 0;
}



////////////////////// PROGRAM ENTRY POINT /////////////////////

int main (int argc, char*argv[])
{
	ProgState st;
	char *hotkey = NULL;
	XFontStruct *font;
	int opt;
	int pen_width = DEFAULT_PEN_WIDTH;
	char *pen_colour = DEFAULT_PEN_COLOUR;
	int outline_width = 1;	

	// default values
	memset(&st, 0, sizeof(st));
	st.hotkey_keycode = st.hotkey_keycode2 = 0;
	asprintf (&st.screenshot_path, "%s/" DEFAULT_SCREENSHOT_PREFIX, getenv("HOME"));
	st.screenshot_delay = 5; // default delay is 5 seconds
	
	// argument processing	
	while ((opt = getopt(argc, argv, "p:k:d:w:c:hn")) != -1) 
	{
		switch (opt) 
		{
			case 'k':
				if (optarg) hotkey = optarg;
				break;
			case 'p':
				if (optarg) st.screenshot_path = optarg;
				break;
			case 'd':
				st.screenshot_delay = atoi (optarg);
				break;
			case 'w':
				pen_width = atoi (optarg);
				if (pen_width<=0) pen_width = DEFAULT_PEN_WIDTH;
				break;
			case 'c':
				pen_colour = optarg;
				break;
			case 'n':
				outline_width = 0;
				break;
			case 'h':
				fprintf (stderr,
"Usage: xscreenshot [options]\n"
"Copyright (C) James Budiono, 2014, 2023. License: GPL version 3.0 or later\n\n"
"Options:\n"
" -k key   : multiple shots mode, use 'key' as hotkey (default none)\n"
" -p path  : path/filename for screenshot (default is $HOME/" DEFAULT_SCREENSHOT_PREFIX ")\n"
" -d delay : the delay (in seconds) for fullscreen screenshot (default 5)\n"
" -w width : width of the selection box in pixels (default 1)\n"
" -c colour: colour of the selection box in pixels (default " DEFAULT_PEN_COLOUR ")\n"
" -n       : disable selection border box outline (default enabled)\n\n"

"If you don't use -k, the program will take one screenshot and exit.\n"
"If you do, the following key combinations controls xscreenshot:\n"
" - hotkey       - toggle region screenshot mode (save to file)\n"
" - Ctrl-hotkey  - toggle region screenshot mode (save to clipboard)\n"
" - Shift-hotkey - delayed fullscreen screenshot with mouse (saves to both)\n"
" - Alt-hotkey   - delayed fullscreen screenshot without mouse (saves to both)\n"
" - Ctrl-Shift-hotkey - clear clipboard\n\n"

"Note: hotkey and Ctrl-hotkey act as a toggle. Press once to enable screenshot-\n"
"taking mode, then take as many as shots as you want. Press again to return to\n"
"normal mouse control. In clipboard mode, on only the most recent shot is kept.\n\n"

"In screenshot-taking mode, press and hold mouse left-button to start the region\n"
"selection, move to choose the region, then release to capture. If you click\n"
"mouse right-button before releasing left-button, the capture is cancelled.\n"
"In single-shot mode, right-click once more to exit the program.\n"
);
				return 0;
		}
	}

	// initialisations
	png_init (0,0);
    st.display = XOpenDisplay(NULL);
    if (st.display == NULL) {
        fprintf(stderr, "Cannot open display\n");
        exit(1);
    }
    st.screen      = DefaultScreen (st.display);
    st.root_window = DefaultRootWindow (st.display);
    st.width       = DisplayWidth(st.display, st.screen);
    st.height      = DisplayHeight(st.display, st.screen);
    st.grab_cursor = XCreateFontCursor (st.display, XC_crosshair);
	font = XLoadQueryFont(st.display, "-*-" MSG_FONTNAME "-*-*-*--" MSG_FONTSIZE "-*-*-*-*-*-*-*");
	
#ifdef USE_OVERLAY
	// create the overlay window to draw selection
	st.wm_delete_window  = XInternAtom(st.display, "WM_DELETE_WINDOW", False);
	st.wm_protocols      = XInternAtom(st.display, "WM_PROTOCOLS", False);    
	create_overlay_window(&st, &st.root_window);
	st.overlay_content   = XCreatePixmap (st.display, st.overlay, st.width, st.height, st.overlay_visinfo.depth);
	st.overlay_shapemask = XCreatePixmap (st.display, st.overlay, st.width, st.height, 1);
	st.bitmap_black      = create_pen (&st, st.overlay_shapemask, "Black", pen_width, font);
	st.bitmap_white      = create_pen (&st, st.overlay_shapemask, "White", pen_width + 2*outline_width, font);
	st.content_blank     = create_pen (&st, st.overlay_content, 0, pen_width, font);
	st.pen               = create_pen (&st, st.overlay_content, pen_colour, pen_width, font);
	clear_drawing(&st);
	repaint(&st);
	show_window(&st, 0);
#else
	st.pen = create_pen (&st, st.root_window, pen_width, font);
#endif

	// for clipboard operation
	st.XA_CLIPBOARD = XInternAtom(st.display, "CLIPBOARD", False);
	st.XA_TARGETS   = XInternAtom(st.display, "TARGETS", False);
	st.XA_MULTIPLE  = XInternAtom(st.display, "MULTIPLE", False);
	st.XA_image_png = XInternAtom(st.display, "image/png", False);
    
    if (hotkey) {
		char *hotkey_name = grab_hotkey (&st, hotkey);
		fprintf (stderr, "Multiple-shot mode, hotkey is: %s.\n", hotkey_name);
	}
    else {
		fprintf(stderr, "Single-shot mode. Right-click to exit.\n");
		// clipboard not supported anyway, must have a running program to do that
		st.save_to_file=1;
		// start selection
		grab_ungrab_mouse (&st, 1);
	}

	// run
	st.select_in_progress = 0;
    event_loop (&st);
    
    // cleanup
    XFreeGC (st.display, st.pen);
	XFlush (st.display);
	XCloseDisplay (st.display);
	return 0;
}

